import base64
from http import HTTPStatus
import sys
from typing import Any, ClassVar, Optional, Type, Union
try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict

from pydantic import BaseModel
import typeguard

from .base_client import BaseClient
from ..exceptions import LagoApiError
from ..services.request import make_headers, make_url, send_get_request
from ..services.response import get_response_data, Response

if sys.version_info >= (3, 9):
    from collections.abc import Mapping, Sequence
else:
    from typing import Mapping, Sequence

typeguard.config.collection_check_strategy = typeguard.CollectionCheckStrategy.ALL_ITEMS


class _ResponseWithPublicKeyInside(TypedDict):
    public_key: str


class WebhookClient(BaseClient):
    API_RESOURCE: ClassVar[str] = 'webhooks'
    RESPONSE_MODEL: ClassVar[Type[BaseModel]] = NotImplemented
    ROOT_NAME: ClassVar[str] = 'webhook'

    def public_key(self) -> bytes:
        api_response: Response = send_get_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, 'json_public_key'),
            ),
            headers=make_headers(api_key=self.api_key),
        )
        response_data: Optional[Union[Mapping[str, Any], Sequence[Any]]] = get_response_data(response=api_response, key=self.ROOT_NAME)

        try:
            checked_response_data: _ResponseWithPublicKeyInside = typeguard.check_type(response_data, _ResponseWithPublicKeyInside)  # type: ignore
        except typeguard.TypeCheckError:
            raise LagoApiError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,  # 500
                url=None,
                response=None,
                detail='Incorrect response data',
                headers=None,
            )

        return base64.b64decode(checked_response_data['public_key'])



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import Invoice, InvoiceMetadata, InvoiceMetadataList


def update_invoice_object():
    metadata = InvoiceMetadata(key='key', value='value')
    metadata_list = InvoiceMetadataList(__root__=[metadata])

    return Invoice(payment_status='failed', metadata=metadata_list)


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    my_data_path = os.path.join(this_dir, 'fixtures/invoice.json')

    with open(my_data_path, 'r') as invoice_response:
        return invoice_response.read()


def mock_collection_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/invoice_index.json')

    with open(data_path, 'r') as invoice_response:
        return invoice_response.read()


def test_valid_update_invoice_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/invoices/5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba',
                       text=mock_response())
        response = client.invoices().update(update_invoice_object(), '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba')

    assert response.lago_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'
    assert response.status == 'finalized'
    assert response.payment_status == 'failed'
    assert response.metadata.__root__[0].key == 'key'
    assert response.metadata.__root__[0].value == 'value'


def test_invalid_update_invoice_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/invoices/5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba',
                       status_code=401,
                       text='')

        with pytest.raises(LagoApiError):
            client.invoices().update(update_invoice_object(), '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba')


def test_valid_find_invoice_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    identifier = '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/invoices/' + identifier, text=mock_response())
        response = client.invoices().find(identifier)

    assert response.lago_id == identifier


def test_invalid_find_invoice_request():
    client = Client(api_key='invalid')
    identifier = '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/invoices/' + identifier, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.invoices().find(identifier)


def test_valid_find_all_invoice_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/invoices', text=mock_collection_response())
        response = client.invoices().find_all()

    assert response['invoices'][0].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1111'
    assert response['meta']['current_page'] == 1


def test_valid_find_all_invoice_request_with_options():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/invoices?per_page=2&page=1', text=mock_collection_response())
        response = client.invoices().find_all({'per_page': 2, 'page': 1})

    assert response['invoices'][1].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1222'
    assert response['meta']['current_page'] == 1


def test_invalid_find_all_invoice_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/invoices', status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.invoices().find_all()


def test_valid_download_invoice_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST',
                        'https://api.getlago.com/api/v1/invoices/5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba/download',
                        text=mock_response())
        response = client.invoices().download('5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba')

    assert response.lago_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'


def test_valid_refresh_invoice_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                        'https://api.getlago.com/api/v1/invoices/5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba/refresh',
                        text=mock_response())
        response = client.invoices().refresh('5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba')

    assert response.lago_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'


def test_valid_finalize_invoice_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                        'https://api.getlago.com/api/v1/invoices/5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba/finalize',
                        text=mock_response())
        response = client.invoices().finalize('5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba')

    assert response.lago_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'


def test_valid_retry_payment_invoice_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri(
            'POST',
            'https://api.getlago.com/api/v1/invoices/5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba/retry_payment',
            text=mock_response())
        response = client.invoices().retry_payment('5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba')

    assert response.lago_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'



from typing import Optional

from pydantic import BaseModel

from .credit import CreditsResponse
from ..base_model import BaseResponseModel


class AppliedCoupon(BaseModel):
    external_customer_id: str
    coupon_code: str
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    percentage_rate: Optional[float]
    frequency: Optional[str]
    frequency_duration: Optional[int]


class AppliedCouponResponse(BaseResponseModel):
    lago_id: str
    lago_coupon_id: str
    coupon_code: str
    status: Optional[str]
    external_customer_id: str
    lago_customer_id: str
    amount_cents: int
    amount_cents_remaining: Optional[int]
    amount_currency: str
    expiration_date: Optional[str]
    created_at: str
    terminated_at: Optional[str]
    percentage_rate: Optional[float]
    frequency: Optional[str]
    frequency_duration: Optional[int]
    frequency_duration_remaining: Optional[int]
    credits: Optional[CreditsResponse]



from typing import Optional

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class Wallet(BaseModel):
    external_customer_id: Optional[str]
    rate_amount: Optional[str]
    name: Optional[str]
    paid_credits: Optional[str]
    granted_credits: Optional[str]
    expiration_at: Optional[str]
    currency: Optional[str]


class WalletResponse(BaseResponseModel):
    lago_id: str
    lago_customer_id: str
    external_customer_id: str
    status: str
    currency: str
    name: Optional[str]
    rate_amount: str
    credits_balance: str
    balance: str
    consumed_credits: str
    created_at: str
    expiration_at: Optional[str]
    last_balance_sync_at: Optional[str]
    last_consumed_credit_at: Optional[str]
    terminated_at: Optional[str]



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin, DestroyCommandMixin, FindAllCommandMixin, FindCommandMixin, UpdateCommandMixin
from ..models.coupon import CouponResponse


class CouponClient(
    CreateCommandMixin[CouponResponse],
    DestroyCommandMixin[CouponResponse],
    FindAllCommandMixin[CouponResponse],
    FindCommandMixin[CouponResponse],
    UpdateCommandMixin[CouponResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'coupons'
    RESPONSE_MODEL: ClassVar[Type[CouponResponse]] = CouponResponse
    ROOT_NAME: ClassVar[str] = 'coupon'



import os

import requests_mock

from lago_python_client.client import Client


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    my_data_path = os.path.join(this_dir, 'fixtures/fee.json')

    with open(my_data_path, 'r') as fee_response:
        return fee_response.read()


def test_valid_find_fee_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    identifier = '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/fees/' + identifier, text=mock_response())
        response = client.fees().find(identifier)

    assert response.lago_id == identifier



from typing import Optional

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class AppliedAddOn(BaseModel):
    external_customer_id: str
    add_on_code: str
    amount_cents: Optional[int]
    amount_currency: Optional[str]


class AppliedAddOnResponse(BaseResponseModel):
    lago_id: str
    lago_add_on_id: str
    add_on_code: str
    external_customer_id: str
    lago_customer_id: str
    amount_cents: int
    amount_currency: str
    created_at: str



from typing import Optional

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class Group(BaseModel):
    lago_id: Optional[str]
    key: Optional[str]
    value: Optional[str]


class GroupResponse(BaseResponseModel):
    lago_id: str
    key: str
    value: str



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models.fee import FeeResponse
from lago_python_client.models.credit_note import Item, Items, CreditNote, CreditNoteUpdate


def credit_note_object():
    item1 = Item(
        fee_id="fee_id_1",
        credit_amount_cents=10,
        refund_amount_cents=10,
    )

    item2 = Item(
        fee="fee_id_2",
        credit_amount_cents=5,
        refund_amount_cents=5,
    )

    return CreditNote(
        lago_id="credit_note_id",
        reason= 'other',
        items= Items(__root__=[item1, item2])
    )


def credit_note_update_object():
    return CreditNoteUpdate(refund_status='pending')


def mock_response():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(current_dir, 'fixtures/credit_note.json')

    with open(data_path, 'r') as credit_note_response:
        return credit_note_response.read()


def mock_collection_response():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(current_dir, 'fixtures/credit_note_index.json')

    with open(data_path, 'r') as credit_notes_response:
        return credit_notes_response.read()


def test_valid_find_credit_note_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    identifier = '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/credit_notes/' + identifier, text=mock_response())
        response = client.credit_notes().find(identifier)

    assert response.lago_id == identifier


def test_invalid_find_invoice_request():
    client = Client(api_key='invalid')
    identifier = '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/credit_notes/' + identifier, status_code=404, text='')

    with pytest.raises(LagoApiError):
        client.credit_notes().find(identifier)


def test_valid_find_all_credit_notes_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/credit_notes', text=mock_collection_response())
        response = client.credit_notes().find_all({'per_page': 2, 'page': 1})

    assert response['credit_notes'][0].lago_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'
    assert response['meta']['current_page'] == 1


def test_valid_download_credit_note_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST',
                        'https://api.getlago.com/api/v1/credit_notes/5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba/download',
                        text=mock_response())
        response = client.credit_notes().download('5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba')

    assert response.lago_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'


def test_valid_create_credit_note_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/credit_notes', text=mock_response())
        response = client.credit_notes().create(credit_note_object())

    assert response.lago_id == "5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba"
    assert response.refund_status == 'pending'


def test_invalid_create_credit_note_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/credit_notes', status_code=422, text='')

        with pytest.raises(LagoApiError):
            client.credit_notes().create(credit_note_object())


def test_valid_update_credit_note_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    credit_note_id = 'credit-note-id'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/credit_notes/' + credit_note_id,
                       text=mock_response())
        response = client.credit_notes().update(credit_note_update_object(), credit_note_id)

    assert response.lago_id == "5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba"
    assert response.refund_status == 'pending'


def test_valid_void_credit_note_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    credit_note_id = 'credit-note-id'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/credit_notes/' + credit_note_id + '/void',
                       text=mock_response())
        response = client.credit_notes().void(credit_note_id)

    assert response.lago_id == "5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba"
    assert response.refund_status == 'pending'



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import BillableMetric, BillableMetricGroup


def billable_metric_object():
    return BillableMetric(
        name='name',
        code='code_first',
        description='desc',
        aggregation_type='sum_agg',
        field_name='amount_sum',
        group=group()
    )


def group():
    return BillableMetricGroup(key='country', values=['france', 'italy', 'spain'])


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/billable_metric.json')

    with open(data_path, 'r') as billable_metric_response:
        return billable_metric_response.read()


def mock_collection_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/billable_metric_index.json')

    with open(data_path, 'r') as billable_metric_response:
        return billable_metric_response.read()


def test_valid_create_billable_metric_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/billable_metrics', text=mock_response())
        response = client.billable_metrics().create(billable_metric_object())

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response.code == 'bm_code'
    assert response.group == group()


def test_invalid_create_billable_metric_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/billable_metrics', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.billable_metrics().create(billable_metric_object())


def test_valid_update_billable_metric_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    code = 'bm_code'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/billable_metrics/' + code,
                       text=mock_response())
        response = client.billable_metrics().update(billable_metric_object(), code)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response.code == code


def test_invalid_update_billable_metric_request():
    client = Client(api_key='invalid')
    code = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/billable_metrics/' + code,
                       status_code=401,
                       text='')

        with pytest.raises(LagoApiError):
            client.billable_metrics().update(billable_metric_object(), code)


def test_valid_find_billable_metric_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    code = 'bm_code'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/billable_metrics/' + code, text=mock_response())
        response = client.billable_metrics().find(code)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response.code == code


def test_invalid_find_billable_metric_request():
    client = Client(api_key='invalid')
    code = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/billable_metrics/' + code, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.billable_metrics().find(code)


def test_valid_destroy_billable_metric_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    code = 'bm_code'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/billable_metrics/' + code, text=mock_response())
        response = client.billable_metrics().destroy(code)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response.code == code


def test_invalid_destroy_billable_metric_request():
    client = Client(api_key='invalid')
    code = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/billable_metrics/' + code, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.billable_metrics().destroy(code)


def test_valid_find_all_billable_metric_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/billable_metrics', text=mock_collection_response())
        response = client.billable_metrics().find_all()

    assert response['billable_metrics'][0].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1000'
    assert response['meta']['current_page'] == 1


def test_valid_find_all_billable_metric_request_with_options():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/billable_metrics?per_page=2&page=1', text=mock_collection_response())
        response = client.billable_metrics().find_all({'per_page': 2, 'page': 1})

    assert response['billable_metrics'][1].lago_id == 'b7ab2926-1de8-4428-9bcd-779314a11111'
    assert response['meta']['current_page'] == 1


def test_invalid_find_all_billable_metric_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/billable_metrics', status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.billable_metrics().find_all()



from datetime import datetime, date, time
from http import HTTPStatus
import sys
from typing import Any, NoReturn, Tuple, Union
from uuid import UUID

from classes import typeclass
import orjson
from requests import Response

from ..exceptions import LagoApiError

if sys.version_info >= (3, 9):
    from collections.abc import Mapping, Sequence
else:
    from typing import Mapping, Sequence

Serializable = Union[
    str, Mapping[Any, Any], Sequence[Any], Tuple[Any], int, float, bool, datetime, date, time, UUID, None,
]  # And dataclass, TypedDict and ndarray
Deserializable = Union[bytes, bytearray, memoryview, str]
DeserializedData = Union[Mapping[str, Any], Sequence[Any], int, float, str, bool, None]


def to_json(data_container: Serializable) -> str:
    """Serialize data into json format."""
    return orjson.dumps(data_container, option=orjson.OPT_NON_STR_KEYS).decode('utf-8')


@typeclass  # type: ignore
def from_json(json_container) -> DeserializedData:
    """Deserialize data from json format."""
    raise TypeError('Type {0} is not supported'.format(type(json_container)))


@from_json.instance(bytes)  # type: ignore
@from_json.instance(bytearray)
@from_json.instance(memoryview)
@from_json.instance(str)
def _from_json_default(json_container: Deserializable) -> DeserializedData:
    """Deserialize json string."""
    try:
        return orjson.loads(json_container)
    except orjson.JSONDecodeError as exc:
        raise LagoApiError(
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,  # 500
            url=None,
            response=None,
            detail=exc.msg,
            headers=None,
        )


@from_json.instance(None)  # type: ignore
def _from_json_none(json_container: None) -> NoReturn:
    """Deserialize json from ``None``."""
    raise LagoApiError(
        status_code=HTTPStatus.INTERNAL_SERVER_ERROR,  # 500
        url=None,
        response=None,
        detail='Input must be bytes, bytearray, memoryview, or str',
        headers=None,
    )


@from_json.instance(Response)  # type: ignore
def _from_json_requests_response_bytes(json_container: Response) -> DeserializedData:
    """Deserialize json from ``requests.Response`` class instances (from content bytes)."""
    return from_json(json_container.content)  # type: ignore



from typing import Optional

from ..base_model import BaseResponseModel


class InvoiceItemResponse(BaseResponseModel):
    type: Optional[str]
    code: Optional[str]
    name: Optional[str]
    lago_item_id: Optional[str]
    item_type: Optional[str]



from typing import List, Optional

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class CustomerBillingConfiguration(BaseModel):
    invoice_grace_period: Optional[int]
    payment_provider: Optional[str]
    provider_customer_id: Optional[str]
    vat_rate: Optional[float]
    sync: Optional[bool]
    sync_with_provider: Optional[bool]
    document_locale: Optional[str]


class Metadata(BaseModel):
    id: Optional[str]
    key: Optional[str]
    value: Optional[str]
    display_in_invoice: Optional[bool]


class MetadataList(BaseModel):
    __root__: List[Metadata]


class Customer(BaseModel):
    external_id: str
    address_line1: Optional[str]
    address_line2: Optional[str]
    city: Optional[str]
    country: Optional[str]
    currency: Optional[str]
    email: Optional[str]
    legal_name: Optional[str]
    legal_number: Optional[str]
    logo_url: Optional[str]
    name: str
    phone: Optional[str]
    state: Optional[str]
    timezone: Optional[str]
    url: Optional[str]
    zipcode: Optional[str]
    metadata: Optional[MetadataList]
    billing_configuration: Optional[CustomerBillingConfiguration]


class CustomerResponse(BaseResponseModel):
    lago_id: str
    external_id: str
    address_line1: Optional[str]
    address_line2: Optional[str]
    city: Optional[str]
    country: Optional[str]
    currency: Optional[str]
    email: Optional[str]
    created_at: str
    legal_name: Optional[str]
    legal_number: Optional[str]
    logo_url: Optional[str]
    name: str
    phone: Optional[str]
    state: Optional[str]
    timezone: Optional[str]
    applicable_timezone: str
    url: Optional[str]
    zipcode: Optional[str]
    metadata: Optional[MetadataList]
    billing_configuration: Optional[CustomerBillingConfiguration]



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin, DestroyCommandMixin, FindAllCommandMixin, FindCommandMixin, UpdateCommandMixin
from ..models.wallet import WalletResponse


class WalletClient(
    CreateCommandMixin[WalletResponse],
    DestroyCommandMixin[WalletResponse],
    FindAllCommandMixin[WalletResponse],
    FindCommandMixin[WalletResponse],
    UpdateCommandMixin[WalletResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'wallets'
    RESPONSE_MODEL: ClassVar[Type[WalletResponse]] = WalletResponse
    ROOT_NAME: ClassVar[str] = 'wallet'



from typing import Optional

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class WalletTransaction(BaseModel):
    wallet_id: Optional[str]
    paid_credits: Optional[str]
    granted_credits: Optional[str]


class WalletTransactionResponse(BaseResponseModel):
    lago_id: str
    lago_wallet_id: str
    status: str
    transaction_type: str
    amount: str
    credit_amount: str
    settled_at: Optional[str]
    created_at: str



from typing import List, Optional

from pydantic import BaseModel

from .credit import CreditsResponse
from .customer import CustomerResponse
from .fee import FeesResponse
from .subscription import SubscriptionsResponse
from ..base_model import BaseResponseModel


# Deprecated: Will be removed in the future
class InvoicePaymentStatusChange(BaseModel):
    payment_status: str


class InvoiceMetadata(BaseModel):
    id: Optional[str]
    key: Optional[str]
    value: Optional[str]


class InvoiceMetadataList(BaseModel):
    __root__: List[InvoiceMetadata]


class Invoice(BaseModel):
    payment_status: Optional[str]
    metadata: Optional[InvoiceMetadataList]


class InvoiceResponse(BaseResponseModel):
    lago_id: str
    sequential_id: int
    issuing_date: Optional[str]
    invoice_type: str
    status: str
    payment_status: str
    amount_cents: int
    amount_currency: str
    vat_amount_cents: int
    vat_amount_currency: str
    total_amount_cents: int
    total_amount_currency: str
    file_url: Optional[str]
    customer: Optional[CustomerResponse]
    subscriptions: Optional[SubscriptionsResponse]
    fees: Optional[FeesResponse]
    credits: Optional[CreditsResponse]
    metadata: Optional[InvoiceMetadataList]



from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class GroupProperties(BaseModel):
    group_id: Optional[str]
    values: Optional[Dict[str, Any]]


class Charge(BaseModel):
    id: Optional[str]
    billable_metric_id: Optional[str]
    charge_model: Optional[str]
    instant: Optional[bool]
    properties: Optional[Dict[str, Any]]
    group_properties: Optional[GroupProperties]


class Charges(BaseModel):
    __root__: List[Charge]


class ChargeResponse(BaseResponseModel):
    lago_id: Optional[str]
    lago_billable_metric_id: Optional[str]
    billable_metric_code: Optional[str]
    charge_model: Optional[str]
    properties: Optional[Dict[str, Any]]
    group_properties: Optional[GroupProperties]


class ChargesResponse(BaseResponseModel):
    __root__: List[ChargeResponse]



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin, DestroyCommandMixin, FindAllCommandMixin, UpdateCommandMixin
from ..models.subscription import SubscriptionResponse


class SubscriptionClient(
    CreateCommandMixin[SubscriptionResponse],
    DestroyCommandMixin[SubscriptionResponse],
    FindAllCommandMixin[SubscriptionResponse],
    UpdateCommandMixin[SubscriptionResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'subscriptions'
    RESPONSE_MODEL: ClassVar[Type[SubscriptionResponse]] = SubscriptionResponse
    ROOT_NAME: ClassVar[str] = 'subscription'



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/webhook.json')

    with open(data_path, 'r') as webhook_response:
        return webhook_response.read()


def test_valid_public_key_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/webhooks/json_public_key', text=mock_response())
        response = client.webhooks().public_key()

    assert response == b'key'


def test_invalid_public_key_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/webhooks/json_public_key', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.webhooks().public_key()



from abc import ABC, abstractmethod
from typing import Type

from pydantic import BaseModel


class BaseClient(ABC):
    """The base class used for each collection client."""

    def __init__(self, base_url: str, api_key: str):
        self.base_url = base_url
        self.api_key = api_key

    @property
    @classmethod
    @abstractmethod
    def API_RESOURCE(cls) -> str:
        """Collection name (required class property) used to build query urls."""
        raise NotImplementedError

    @property
    @classmethod
    @abstractmethod
    def RESPONSE_MODEL(cls) -> Type[BaseModel]:
        """Response model (required class property) used to prepare response."""
        raise NotImplementedError

    @property
    @classmethod
    @abstractmethod
    def ROOT_NAME(cls) -> str:
        """The resource key (required class property), used to access the response data."""
        raise NotImplementedError



from typing import List, Optional

from .invoice_item import InvoiceItemResponse
from ..base_model import BaseModel, BaseResponseModel

class Fee(BaseModel):
    payment_status: Optional[str]

class FeeResponse(BaseResponseModel):
    lago_id: Optional[str]
    lago_group_id: Optional[str]
    lago_invoice_id: Optional[str]
    external_subscription_id: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    vat_amount_cents: Optional[int]
    vat_amount_currency: Optional[str]
    total_amount_cents: Optional[int]
    total_amount_currency: Optional[str]
    units: Optional[float]
    events_count: Optional[int]
    payment_status: Optional[str]
    created_at: Optional[str]
    succeeded_at: Optional[str]
    failed_at: Optional[str]
    refunded_at: Optional[str]
    from_date: Optional[str]
    to_date: Optional[str]

    item: Optional[InvoiceItemResponse]


class FeesResponse(BaseResponseModel):
    __root__: List[FeeResponse]



"""Test request services."""
import requests

from lago_python_client.services.request import make_headers, make_url, send_delete_request, send_get_request, send_post_request, send_put_request
from lago_python_client.version import LAGO_VERSION


def test_make_headers():
    """Make headers."""
    # Give api_key
    api_key: str = "test"
    # When service is applied
    result = make_headers(api_key=api_key)
    # Then
    assert result == {
        'Content-type': 'application/json',
        'Authorization': "Bearer test",
        'User-agent': 'Lago Python v{version}'.format(version=LAGO_VERSION),
    }


def test_make_url():
    """Make url."""
    # Given
    api_url = 'https://api.getlago.com/api/v1/'
    some_path_parts = ('team', 'anhtho', 'congratulate')
    query_name_value = {
        'message': "The future belongs to those who believe in the beauty of their dreams. Happy International Women's Day!",
        'day': 8,
    }

    # When service is applied
    result = make_url(origin=api_url, path_parts=some_path_parts, query_pairs=query_name_value)
    # Then
    assert result == 'https://api.getlago.com/api/v1/team/anhtho/congratulate?message=The+future+belongs+to+those+who+believe+in+the+beauty+of+their+dreams.+Happy+International+Women%27s+Day%21&day=8'


def test_make_url_no_query():
    """Make url without query name-value pairs."""
    # Given
    api_url = 'https://api.getlago.com/api/v1/'
    some_path_parts = ('hello', )

    # When service is applied
    result = make_url(origin=api_url, path_parts=some_path_parts)
    # Then
    assert result == 'https://api.getlago.com/api/v1/hello'


def test_send_delete_request():
    """Ensure `send_delete_request` service use requests."""
    assert send_delete_request == requests.delete


def test_send_get_request():
    """Ensure `send_get_request` service use requests."""
    assert send_get_request == requests.get


def test_send_post_request():
    """Ensure `send_post_request` service use requests."""
    assert send_post_request == requests.post


def test_send_put_request():
    """Ensure `send_put_request` service use requests."""
    assert send_put_request == requests.put



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin, DestroyCommandMixin, FindAllCommandMixin, FindCommandMixin, UpdateCommandMixin
from ..models.add_on import AddOnResponse


class AddOnClient(
    CreateCommandMixin[AddOnResponse],
    DestroyCommandMixin[AddOnResponse],
    FindAllCommandMixin[AddOnResponse],
    FindCommandMixin[AddOnResponse],
    UpdateCommandMixin[AddOnResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'add_ons'
    RESPONSE_MODEL: ClassVar[Type[AddOnResponse]] = AddOnResponse
    ROOT_NAME: ClassVar[str] = 'add_on'



import sys
from typing import Any, Generic, Optional, Type, TypeVar, Union
try:
    from typing import Protocol
except ImportError:  # Python 3.7
    from typing_extensions import Protocol  # type: ignore

from pydantic import BaseModel

from .services.json import to_json
from .services.request import make_headers, make_url, send_delete_request, send_get_request, send_post_request, send_put_request
from .services.response import get_response_data, prepare_index_response, prepare_object_response, Response

if sys.version_info >= (3, 9):
    from collections.abc import Mapping
else:
    from typing import Mapping

_PM = TypeVar("_PM", covariant=True)
_M = TypeVar("_M", bound=BaseModel)


class _ClientMixin(Protocol[_PM]):
    @property
    def API_RESOURCE(self) -> str: ...
    @property
    def RESPONSE_MODEL(self) -> Type[_PM]: ...
    @property
    def ROOT_NAME(self) -> str: ...
    @property
    def api_key(self) -> str: ...
    @property
    def base_url(self) -> str: ...


class CreateCommandMixin(Generic[_M]):
    """Client mixin with `create` command."""

    def create(self: _ClientMixin[_M], input_object: BaseModel) -> Union[Optional[_M], bool]:
        """Execute `create` command."""
        # Send request and save response
        api_response: Response = send_post_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, ),
            ),
            data=to_json({
                self.ROOT_NAME: input_object.dict(),
            }),
            headers=make_headers(api_key=self.api_key),
        )

        # Process response data
        response_data = get_response_data(response=api_response, key=self.ROOT_NAME)
        if not response_data:
            return True  # TODO: should return None

        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=response_data,
        )


class DestroyCommandMixin(Generic[_M]):
    """Client mixin with `destroy` command."""

    def destroy(self: _ClientMixin[_M], resource_id: str) -> BaseModel:
        """Execute `destroy` command."""
        # Send request and save response
        api_response: Response = send_delete_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, resource_id),
            ),
            headers=make_headers(api_key=self.api_key),
        )

        # Process response data
        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response, key=self.ROOT_NAME),
        )


class FindAllCommandMixin(Generic[_M]):
    """Client mixin with `find_all` command."""

    def find_all(self: _ClientMixin[_M], options: Mapping[str, Union[int, str]] = {}) -> Mapping[str, Any]:
        """Execute `find all` command."""
        # Send request and save response
        api_response: Response = send_get_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, ),
                query_pairs=options,
            ),
            headers=make_headers(api_key=self.api_key),
        )

        # Process response data
        return prepare_index_response(
            api_resource=self.API_RESOURCE,
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response),
        )


class FindCommandMixin(Generic[_M]):
    """Client mixin with `find` command."""

    def find(self: _ClientMixin[_M], resource_id: str, params: Mapping[str, str] = {}) -> _M:
        """Execute `find` command."""
        # Send request and save response
        api_response: Response = send_get_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, resource_id),
            ),
            data=to_json(params) if params else None,
            headers=make_headers(api_key=self.api_key),
        )

        # Process response data
        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response, key=self.ROOT_NAME),
        )


class UpdateCommandMixin(Generic[_M]):
    """Client mixin with `update` command."""

    def update(self: _ClientMixin[_M], input_object: BaseModel, identifier: Optional[str] = None) -> _M:
        """Execute `update` command."""
        # Send request and save response
        api_response: Response = send_put_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, identifier) if identifier else (self.API_RESOURCE, ),
            ),
            data=to_json({
                self.ROOT_NAME: input_object.dict(exclude_none=True),
            }),
            headers=make_headers(api_key=self.api_key),
        )

        # Process response data
        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response, key=self.ROOT_NAME),
        )



from .add_on_client import AddOnClient
from .applied_add_on_client import AppliedAddOnClient
from .applied_coupon_client import AppliedCouponClient
from .billable_metric_client import BillableMetricClient
from .coupon_client import CouponClient
from .credit_note_client import CreditNoteClient
from .customer_client import CustomerClient
from .event_client import EventClient
from .fee_client import FeeClient
from .group_client import GroupClient
from .invoice_client import InvoiceClient
from .organization_client import OrganizationClient
from .plan_client import PlanClient
from .subscription_client import SubscriptionClient
from .wallet_client import WalletClient
from .wallet_transaction_client import WalletTransactionClient
from .webhook_client import WebhookClient



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import UpdateCommandMixin
from ..models.organization import OrganizationResponse


class OrganizationClient(
    UpdateCommandMixin[OrganizationResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'organizations'
    RESPONSE_MODEL: ClassVar[Type[OrganizationResponse]] = OrganizationResponse
    ROOT_NAME: ClassVar[str] = 'organization'



from typing import ClassVar, Optional, Type, Union

from .base_client import BaseClient
from ..mixins import FindAllCommandMixin, FindCommandMixin, UpdateCommandMixin
from ..models.invoice import InvoiceResponse
from ..services.request import make_headers, make_url, send_post_request, send_put_request
from ..services.response import get_response_data, prepare_object_response, Response


class InvoiceClient(
    FindCommandMixin[InvoiceResponse],
    FindAllCommandMixin[InvoiceResponse],
    UpdateCommandMixin[InvoiceResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'invoices'
    RESPONSE_MODEL: ClassVar[Type[InvoiceResponse]] = InvoiceResponse
    ROOT_NAME: ClassVar[str] = 'invoice'

    def download(self, resource_id: str) -> Union[Optional[InvoiceResponse], bool]:
        api_response: Response = send_post_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, resource_id, 'download'),
            ),
            headers=make_headers(api_key=self.api_key),
        )

        response_data = get_response_data(response=api_response, key=self.ROOT_NAME)
        if not response_data:
            return True  # TODO: should return None

        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=response_data,
        )

    def retry_payment(self, resource_id: str) -> InvoiceResponse:
        api_response: Response = send_post_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, resource_id, 'retry_payment'),
            ),
            headers=make_headers(api_key=self.api_key),
        )

        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response, key=self.ROOT_NAME),
        )

    def refresh(self, resource_id: str) -> InvoiceResponse:
        api_response: Response = send_put_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, resource_id, 'refresh'),
            ),
            headers=make_headers(api_key=self.api_key),
        )

        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response, key=self.ROOT_NAME),
        )

    def finalize(self, resource_id: str) -> InvoiceResponse:
        api_response: Response = send_put_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, resource_id, 'finalize'),
            ),
            headers=make_headers(api_key=self.api_key),
        )

        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response, key=self.ROOT_NAME),
        )



from typing import Any, List, Optional

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class LimitationConfiguration(BaseModel):
    plan_codes: Optional[List[Any]]


class Coupon(BaseModel):
    name: Optional[str]
    code: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    expiration: Optional[str]
    expiration_at: Optional[str]
    percentage_rate: Optional[float]
    coupon_type: Optional[str]
    reusable: Optional[bool]
    frequency: Optional[str]
    frequency_duration: Optional[int]
    applies_to: Optional[LimitationConfiguration]


class CouponResponse(BaseResponseModel):
    lago_id: str
    name: str
    code: str
    amount_cents: int
    amount_currency: str
    created_at: str
    expiration: str
    expiration_at: Optional[str]
    percentage_rate: Optional[float]
    coupon_type: Optional[str]
    reusable: Optional[bool]
    frequency: Optional[str]
    frequency_duration: Optional[int]
    plan_codes: Optional[List[Any]]
    limited_plans: Optional[bool]



from typing import List, Optional

from pydantic import BaseModel

from .invoice_item import InvoiceItemResponse
from ..base_model import BaseResponseModel


class InvoiceShortDetails(BaseModel):
    lago_id: Optional[str]
    payment_status: Optional[str]


class CreditResponse(BaseResponseModel):
    lago_id: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    item: Optional[InvoiceItemResponse]
    invoice: Optional[InvoiceShortDetails]


class CreditsResponse(BaseResponseModel):
    __root__: List[CreditResponse]



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import Customer, CustomerBillingConfiguration, Metadata, MetadataList


def create_customer():
    metadata = Metadata(
        display_in_invoice=True,
        key='key',
        value='value'
    )
    metadata_list = MetadataList(__root__=[metadata])

    return Customer(
        external_id='5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba',
        name='Gavin Belson',
        currency='EUR',
        billing_configuration=CustomerBillingConfiguration(
            invoice_grace_period=3,
            payment_provider='stripe',
            provider_customer_id='cus_12345',
            sync_with_provider=True,
            vat_rate=12.5,
            document_locale="fr"
        ),
        metadata=metadata_list
    )


def mock_response(mock='customer'):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    my_data_path = os.path.join(this_dir, 'fixtures/' + mock + '.json')

    with open(my_data_path, 'r') as customer_response:
        return customer_response.read()


def test_valid_create_customers_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/customers', text=mock_response())
        response = client.customers().create(create_customer())

    assert response.external_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'
    assert response.name == 'Gavin Belson'
    assert response.email == 'dinesh@piedpiper.test'
    assert response.currency == 'EUR'
    assert response.billing_configuration.invoice_grace_period == 3
    assert response.billing_configuration.payment_provider == 'stripe'
    assert response.billing_configuration.provider_customer_id == 'cus_12345'
    assert response.billing_configuration.sync_with_provider == True
    assert response.billing_configuration.vat_rate == 12.5
    assert response.billing_configuration.document_locale == "fr"
    assert response.metadata.__root__[0].key == 'key'
    assert response.metadata.__root__[0].value == 'value'


def test_invalid_create_customers_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/customers', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.customers().create(create_customer())


def test_valid_current_usage():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET',
                       'https://api.getlago.com/api/v1/customers/external_customer_id/current_usage?external_subscription_id=123',
                       text=mock_response('customer_usage'))
        response = client.customers().current_usage('external_customer_id', '123')

    assert response.from_datetime == '2022-07-01T00:00:00Z'
    assert len(response.charges_usage) == 1
    assert response.charges_usage[0].units == 1.0


def test_invalid_current_usage():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET',
                       'https://api.getlago.com/api/v1/customers/invalid_customer/current_usage?external_subscription_id=123',
                       status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.customers().current_usage('invalid_customer', '123')


def test_valid_portal_url():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET',
                       'https://api.getlago.com/api/v1/customers/external_customer_id/portal_url',
                       text=mock_response('customer_portal_url'))
        response = client.customers().portal_url('external_customer_id')

    assert response == "https://app.lago.dev/customer-portal/eyJfcmFpbHMiOnsibWVzc2FnZSI6IkJBaEpJaWt3WkdGbE1qWmxZUzFqWlRnekxUUTJZell0T1dRNFl5MHdabVF4TURabFlqY3dNVElHT2daRlZBPT0iLCJleHAiOiIyMDIzLTAzLTIzVDIzOjAzOjAwLjM2NloiLCJwdXIiOm51bGx9fQ==--7128c6e541adc7b4c14249b1b18509f92e652d17"


def test_invalid_portal_url():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET',
                       'https://api.getlago.com/api/v1/customers/invalid_customer/portal_url',
                       status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.customers().portal_url('invalid_customer')


def test_valid_destroy_customer_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    external_id = '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/customers/' + external_id, text=mock_response())
        response = client.customers().destroy(external_id)

    assert response.lago_id == '99a6094e-199b-4101-896a-54e927ce7bd7'
    assert response.external_id == external_id


def test_invalid_destroy_customer_request():
    client = Client(api_key='invalid')
    external_id = 'external_id'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/customers/' + external_id, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.customers().destroy(external_id)



from http import HTTPStatus
import sys
from typing import Any, Optional, Set, Type, TypeVar, Union
try:
    from typing import assert_never
except ImportError:  # Python 3.7, 3.8, 3.9, 3.10
    from typing_extensions import assert_never
try:
    from typing import Final
except ImportError:  # Python 3.7
    from typing_extensions import Final  # type: ignore

from pydantic import BaseModel
from requests import Response as Response  # not a typo! implicit reexport
import typeguard

from ..exceptions import LagoApiError
from ..services.json import DeserializedData, from_json

if sys.version_info >= (3, 9):
    from collections.abc import Mapping, Sequence
else:
    from typing import Mapping, Sequence

_M = TypeVar("_M", bound=BaseModel)

typeguard.config.collection_check_strategy = typeguard.CollectionCheckStrategy.ALL_ITEMS

RESPONSE_SUCCESS_CODES: Final[Set[int]] = {
    HTTPStatus.OK,  # 200
    HTTPStatus.CREATED,  # 201
    HTTPStatus.ACCEPTED,  # 202
    HTTPStatus.NO_CONTENT,  # 204
}

_MappingOrSequence = Union[Mapping[str, Any], Sequence[Any]]


def _is_status_code_successful(response: Response) -> bool:
    """Check status code."""
    return response.status_code in RESPONSE_SUCCESS_CODES


def _is_content_exists(response: Response) -> bool:
    """Check content is not empty."""
    return bool(response.content)


def verify_response(response: Response) -> Optional[Response]:
    """Verify response."""
    if not _is_status_code_successful(response):
        response_data: Any = from_json(response)  # type: ignore
        raise LagoApiError(
            status_code=response.status_code,
            url=response.request.url,
            response=response_data,
            detail=getattr(response_data, 'error', None),
            headers=response.headers,
        )

    if not _is_content_exists(response):
        return None

    return response


def get_response_data(*, response: Response, key: str = '') -> Optional[_MappingOrSequence]:
    """Return verified and unpacked response data."""
    response_or_None: Optional[Response] = verify_response(response)
    if not response_or_None:
        return None
    deserialized_data: DeserializedData = from_json(response_or_None)  # type: ignore

    # Ensure deserialized_data has correct type: sequence or mapping or raise LagoApiError
    try:
        mapping_or_sequence_data = typeguard.check_type(deserialized_data, _MappingOrSequence)
    except typeguard.TypeCheckError as exc:
        raise LagoApiError(
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,  # 500
            url=None,
            response=None,
            detail=str(exc),
            headers=None,
        )

    if isinstance(mapping_or_sequence_data, Mapping):
        return mapping_or_sequence_data.get(key, {}) if key else mapping_or_sequence_data
    elif isinstance(mapping_or_sequence_data, Sequence):
        return mapping_or_sequence_data
    else:
        assert_never(mapping_or_sequence_data)


def prepare_object_response(response_model: Type[_M], data: Optional[_MappingOrSequence]) -> _M:
    """Return single object response - Pydantic model instance with provided data."""
    if not data:
        raise LagoApiError(
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,  # 500
            url=None,
            response=None,
            detail='Data is required to create instance of {model}'.format(model=response_model),
            headers=None,
        )

    return response_model.parse_obj(data)


def prepare_index_response(api_resource: str, response_model: Type[_M], data: Optional[_MappingOrSequence]) -> Mapping[str, Any]:
    """Return index response with meta based on mapping data object."""
    # Ensure deserialized_data has correct type: mapping with mapping or sequence inside or raise LagoApiError

    try:
        response_data: Mapping[str, _MappingOrSequence] = typeguard.check_type(data, Mapping[str, _MappingOrSequence])
    except typeguard.TypeCheckError as exc:
        raise LagoApiError(
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,  # 500
            url=None,
            response=None,
            detail=str(exc),
            headers=None,
        )

    return {
        api_resource: [
            prepare_object_response(response_model=response_model, data=el)
            for el in response_data[api_resource]
        ],
        'meta': response_data['meta'],
    }


def prepare_object_list_response(
        api_resource: str,
        response_model: Type[_M],
        data: Optional[Union[Mapping[str, object], Sequence[object]]],
) -> Mapping[str, Any]:
    """Return response based on sequence of data objects."""
    # The only usage - ``WalletTransactionClient.create``
    # Ensure deserialized_data has correct type: sequence with mapping or sequence inside or raise LagoApiError
    try:
        response_data: Sequence[_MappingOrSequence] = typeguard.check_type(data, Sequence[_MappingOrSequence])
    except typeguard.TypeCheckError as exc:
        raise LagoApiError(
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,  # 500
            url=None,
            response=None,
            detail=str(exc),
            headers=None,
        )

    return {
        api_resource: [
            prepare_object_response(response_model=response_model, data=el)
            for el in response_data
        ],
    }



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import WalletTransaction


def wallet_transaction_object():
    return WalletTransaction(
        wallet_id='123',
        paid_credits='10',
        granted_credits='10'
    )


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/wallet_transaction.json')

    with open(data_path, 'r') as wallet_transaction_response:
        return wallet_transaction_response.read()


def mock_collection_response():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(current_dir, 'fixtures/wallet_transaction_index.json')

    with open(data_path, 'r') as wallet_transaction_index_response:
        return wallet_transaction_index_response.read()


def test_valid_create_wallet_transaction_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/wallet_transactions', text=mock_response())
        response = client.wallet_transactions().create(wallet_transaction_object())

    assert response['wallet_transactions'][0].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1111'
    assert response['wallet_transactions'][1].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1222'


def test_invalid_create_wallet_transaction_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/wallet_transactions', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.wallet_transactions().create(wallet_transaction_object())


def test_valid_find_all_groups_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri(
            'GET',
            'https://api.getlago.com/api/v1/wallets/555/wallet_transactions',
            text=mock_collection_response()
        )
        response = client.wallet_transactions().find_all('555')

    assert response['wallet_transactions'][0].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1111'
    assert response['meta']['current_page'] == 1



from lago_python_client.client import Client


def test_base_url_when_api_url_is_given():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d',
                    api_url='https://example.com/')

    assert client.base_api_url == 'https://example.com/api/v1/'


def test_base_url_when_api_url_is_not_given():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    assert client.base_api_url == 'https://api.getlago.com/api/v1/'



from typing import List

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class Metric(BaseModel):
    lago_id: str
    name: str
    code: str
    aggregation_type: str


class ChargeObject(BaseModel):
    lago_id: str
    charge_model: str


class ChargeUsage(BaseModel):
    units: float
    amount_cents: int
    amount_currency: str
    charge: ChargeObject
    billable_metric: Metric


class CustomerUsageResponse(BaseResponseModel):
    from_datetime: str
    to_datetime: str
    issuing_date: str
    amount_cents: int
    amount_currency: str
    total_amount_cents: int
    total_amount_currency: str
    vat_amount_cents: int
    vat_amount_currency: str
    charges_usage: List[ChargeUsage]



from urllib.parse import urljoin

from .clients.applied_add_on_client import AppliedAddOnClient
from .clients.applied_coupon_client import AppliedCouponClient
from .clients.billable_metric_client import BillableMetricClient
from .clients.coupon_client import CouponClient
from .clients.group_client import GroupClient
from .clients.credit_note_client import CreditNoteClient
from .clients.plan_client import PlanClient
from .clients.add_on_client import AddOnClient
from .clients.organization_client import OrganizationClient
from .clients.subscription_client import SubscriptionClient
from .clients.customer_client import CustomerClient
from .clients.invoice_client import InvoiceClient
from .clients.event_client import EventClient
from .clients.fee_client import FeeClient
from .clients.webhook_client import WebhookClient
from .clients.wallet_client import WalletClient
from .clients.wallet_transaction_client import WalletTransactionClient
from .functools_ext import callable_cached_property

try:
    from typing import Final
except ImportError:  # Python 3.7
    from typing_extensions import Final  # type: ignore


class Client:
    BASE_URL: Final[str] = 'https://api.getlago.com/'
    API_PATH: Final[str] = 'api/v1/'

    def __init__(self, api_key: str = '', api_url: str = '') -> None:
        self.api_key: str = api_key
        self.api_url: str = api_url

    @property
    def base_api_url(self) -> str:
        return urljoin(self.api_url if self.api_url else Client.BASE_URL, Client.API_PATH)

    @callable_cached_property
    def events(self) -> EventClient:
        return EventClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def fees(self) -> FeeClient:
        return FeeClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def groups(self) -> GroupClient:
        return GroupClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def subscriptions(self) -> SubscriptionClient:
        return SubscriptionClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def credit_notes(self) -> CreditNoteClient:
        return CreditNoteClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def customers(self) -> CustomerClient:
        return CustomerClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def invoices(self) -> InvoiceClient:
        return InvoiceClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def applied_coupons(self) -> AppliedCouponClient:
        return AppliedCouponClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def applied_add_ons(self) -> AppliedAddOnClient:
        return AppliedAddOnClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def billable_metrics(self) -> BillableMetricClient:
        return BillableMetricClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def coupons(self) -> CouponClient:
        return CouponClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def plans(self) -> PlanClient:
        return PlanClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def add_ons(self) -> AddOnClient:
        return AddOnClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def organizations(self) -> OrganizationClient:
        return OrganizationClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def webhooks(self) -> WebhookClient:
        return WebhookClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def wallets(self) -> WalletClient:
        return WalletClient(self.base_api_url, self.api_key)

    @callable_cached_property
    def wallet_transactions(self) -> WalletTransactionClient:
        return WalletTransactionClient(self.base_api_url, self.api_key)



"""Test response services."""
from copy import deepcopy

from pydantic import BaseModel
import pytest
from requests import Request, Response

from lago_python_client.exceptions import LagoApiError
from lago_python_client.services.response import (
    Response as ServiceResponse,
    RESPONSE_SUCCESS_CODES, _is_status_code_successful, _is_content_exists, verify_response,
    get_response_data, prepare_object_list_response, prepare_index_response, prepare_object_response,
)


def test_Response():
    """Ensure Response object is the same as ``request.Response``."""
    assert ServiceResponse == Response


def test_success_status_codes_exists():
    """Ensure ``RESPONSE_SUCCESS_CODES`` is set."""
    assert len(RESPONSE_SUCCESS_CODES) >= 4


def test_is_status_code_successful():
    """Check status code is successful."""
    # Given instanse of ``requests.Response`` class with successful status code
    response = Response()
    response._content_consumed = True
    response._content = b'{"a":{"b":"c"}}'
    response.status_code = 200
    # ... and with error status code
    response404 = deepcopy(response)
    response404._content = b''
    response404.status_code = 404

    # When check helper function is applied
    # Then verification passes.
    assert _is_status_code_successful(response) is True
    # ... or not
    assert _is_status_code_successful(response404) is False


def test_is_content_exists():
    """Check content exists."""
    # Given instanse of ``requests.Response`` class with content
    response = Response()
    response._content_consumed = True
    response._content = b'{"a":{"b":"c"}}'
    response.status_code = 200
    # ... and with empty content
    response204 = deepcopy(response)
    response204._content = b''
    response204.status_code = 204

    # When check helper function is applied
    # Then verification passes.
    assert _is_content_exists(response) is True
    # ... or not
    assert _is_content_exists(response204) is False


def test_verify_response():
    """Verify response."""
    # Given instanse of ``requests.Response`` class with successful status code and content
    response = Response()
    response._content_consumed = True
    response._content = b'{"a":{"b":"c"}}'
    response.status_code = 200
    # ... and with successful status code and empty content
    response204 = deepcopy(response)
    response204._content = b''
    response204.status_code = 204
    # ... and with error status code
    response404 = deepcopy(response)
    response404.status_code = 404
    response404.request = Request(url='')

    # When service is applied
    # Then
    assert verify_response(response) == response
    assert verify_response(response204) is None
    with pytest.raises(LagoApiError):
        verify_response(response404)


def test_get_response_data():
    """Get response data."""
    # Given instanse of ``requests.Response`` class with successful status code and content
    response = Response()
    response._content_consumed = True
    response._content = b'{"a":{"b":"c"}}'
    response.status_code = 200
    # ... and with successful status code and content with sequence inside
    response200_2 = deepcopy(response)
    response200_2._content = b'[{"a": "b"}]'
    # ... and with successful status code and empty content
    response204 = deepcopy(response)
    response204._content = b''
    response204.status_code = 204
    # ... and with error status code
    response404 = deepcopy(response)
    response404.status_code = 404
    response404.request = Request(url='')

    # When service is applied
    # Then
    assert get_response_data(response=response) == {
        'a': {
            'b': 'c',
        },
    }
    assert get_response_data(response=response, key='a') == {
        'b': 'c',
    }
    assert get_response_data(response=response200_2) == [
        {
            'a': 'b',
        },
    ]
    assert get_response_data(response=response204) is None
    with pytest.raises(LagoApiError):
        get_response_data(response=response404)


class SomeHumanModel(BaseModel):
    name: str


def test_prepare_object_response():
    """Verify ``prepare_object_response`` service returns Pydantic model instance."""
    # Given Pydantic model and some data
    data = {'name': 'Aurelia'}

    # When service is applied
    result = prepare_object_response(response_model=SomeHumanModel, data=data)
    # Then
    assert SomeHumanModel(**data) == result


def test_prepare_index_response():
    """Verify ``prepare_index_response`` service returns valid mapping object."""
    # Given Pydantic model and some data
    data = {
        'human': [
            {'name': 'Aurelia'},
            {'name': 'Aleksandra'},
            {'name': 'John'},
        ],
        'meta': {
            'something': 'is here'
        },
    }

    # When service is applied
    result = prepare_index_response(api_resource='human', response_model=SomeHumanModel, data=data)
    # Then
    assert SomeHumanModel(**data['human'][0]) in result['human']
    assert len(result['human']) == 3
    assert 'meta' in result


def test_prepare_object_list_response():
    """Verify ``prepare_object_list_response`` service returns valid mapping object."""
    # Given Pydantic model and some data
    data = [
        {'name': 'Aurelia'},
        {'name': 'Aleksandra'},
        {'name': 'John'},
    ]

    # When service is applied
    result = prepare_object_list_response(api_resource='human', response_model=SomeHumanModel, data=data)
    # Then
    assert SomeHumanModel(**data[0]) == result['human'][0]
    assert len(result['human']) == 3



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import FindCommandMixin, FindAllCommandMixin, UpdateCommandMixin
from ..models.fee import FeeResponse


class FeeClient(
    FindCommandMixin[FeeResponse],
    FindAllCommandMixin[FeeResponse],
    UpdateCommandMixin[FeeResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'fees'
    RESPONSE_MODEL: ClassVar[Type[FeeResponse]] = FeeResponse
    ROOT_NAME: ClassVar[str] = 'fee'



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import Subscription


def create_subscription():
    return Subscription(external_customer_id='5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba', plan_code='eartha lynch',
                        external_id='code', billing_time='anniversary', subscription_date='2022-04-29')


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    my_data_path = os.path.join(this_dir, 'fixtures/subscription.json')

    with open(my_data_path, 'r') as subscription_response:
        return subscription_response.read()


def mock_collection_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/subscription_index.json')

    with open(data_path, 'r') as subscription_response:
        return subscription_response.read()


def test_valid_create_subscriptions_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/subscriptions', text=mock_response())
        response = client.subscriptions().create(create_subscription())

    assert response.external_customer_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'
    assert response.status == 'active'
    assert response.plan_code == 'eartha lynch'
    assert response.billing_time == 'anniversary'
    assert response.subscription_date == '2022-04-29'


def test_invalid_create_subscriptions_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/subscriptions', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.subscriptions().create(create_subscription())


def test_valid_update_subscription_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    identifier = 'sub_id'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/subscriptions/' + identifier,
                       text=mock_response())
        response = client.subscriptions().update(Subscription(name='name'), identifier)

    assert response.external_customer_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'
    assert response.status == 'active'
    assert response.plan_code == 'eartha lynch'
    assert response.billing_time == 'anniversary'
    assert response.subscription_date == '2022-04-29'


def test_invalid_update_subscription_request():
    client = Client(api_key='invalid')
    identifier = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/subscriptions/' + identifier,
                       status_code=401,
                       text='')

        with pytest.raises(LagoApiError):
            client.subscriptions().update(Subscription(name='name'), identifier)


def test_valid_destroy_subscription_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    identifier = 'sub_id'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/subscriptions/' + identifier, text=mock_response())
        response = client.subscriptions().destroy(identifier)

    assert response.external_customer_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'
    assert response.status == 'active'
    assert response.plan_code == 'eartha lynch'


def test_invalid_destroy_subscription_request():
    client = Client(api_key='invalid')
    identifier = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/subscriptions/' + identifier,
                       status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.subscriptions().destroy(identifier)


def test_valid_find_all_subscription_request_with_options():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/subscriptions?external_customer_id=123',
                       text=mock_collection_response())
        response = client.subscriptions().find_all({'external_customer_id': '123'})

    assert response['subscriptions'][0].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response['meta']['current_page'] == 1


def test_invalid_find_all_subscription_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/subscriptions', status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.subscriptions().find_all()



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models.event import BatchEvent, Event


def create_event():
    return Event(external_customer_id='5eb02857-a71e-4ea2-bcf9-57d8885990ba', code='123', transaction_id='123')


def create_batch_event():
    return BatchEvent(external_subscription_ids=['88u02857-a71e-4ea2-bcf9-57d8885990ba'], code='123', transaction_id='123')


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/event.json')

    with open(data_path, 'r') as event_response:
        return event_response.read()

def mock_fees_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/fees.json')

    with open(data_path, 'r') as fees_response:
        return fees_response.read()


def test_valid_create_events_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/events', text='')
        response = client.events().create(create_event())

    assert response == True


def test_invalid_create_events_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/events', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.events().create(create_event())


def test_valid_create_batch_events_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/events/batch', text='')
        response = client.events().batch_create(create_batch_event())

    assert response == True


def test_invalid_create_batch_events_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/events/batch', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.events().batch_create(create_batch_event())


def test_valid_find_event_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    event_id = '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/events/' + event_id, text=mock_response())
        response = client.events().find(event_id)

    assert response.lago_id == event_id


def test_invalid_find_events_request():
    client = Client(api_key='invalid')
    event_id = 'INVALID_EVENT'

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/events', status_code=401, text='')
        m.register_uri('GET', 'https://api.getlago.com/api/v1/events/' + event_id, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.events().find(event_id)


def test_valid_estimate_fees_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/events/estimate_fees', text=mock_fees_response())
        response = client.events().estimate_fees(create_event())

    assert response['fees'][0].item.type == 'instant_charge'


def test_invalid_estimate_fees_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/events/estimate_fees', status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.events().estimate_fees(create_event())



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import Coupon, LimitationConfiguration


def coupon_object():
    return Coupon(
        name='name',
        code='code_first',
        amount_cents=1000,
        amount_currency='EUR',
        expiration='no_expiration',
        expiration_at="2022-08-08T23:59:59Z",
        coupon_type="fixed_amount",
        reusable=False,
        frequency="once",
        applies_to=LimitationConfiguration(plan_codes=['plan1'])
    )


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/coupon.json')

    with open(data_path, 'r') as coupon_response:
        return coupon_response.read()


def mock_collection_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/coupon_index.json')

    with open(data_path, 'r') as coupon_response:
        return coupon_response.read()


def test_valid_create_coupon_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/coupons', text=mock_response())
        response = client.coupons().create(coupon_object())

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response.code == 'coupon_code'


def test_invalid_create_coupon_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/coupons', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.coupons().create(coupon_object())


def test_valid_update_coupon_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    code = 'coupon_code'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/coupons/' + code,
                       text=mock_response())
        response = client.coupons().update(coupon_object(), code)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response.code == code


def test_invalid_update_coupon_request():
    client = Client(api_key='invalid')
    code = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/coupons/' + code,
                       status_code=401,
                       text='')

        with pytest.raises(LagoApiError):
            client.coupons().update(coupon_object(), code)


def test_valid_find_coupon_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    code = 'coupon_code'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/coupons/' + code, text=mock_response())
        response = client.coupons().find(code)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response.code == code


def test_invalid_find_coupon_request():
    client = Client(api_key='invalid')
    code = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/coupons/' + code, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.coupons().find(code)


def test_valid_destroy_coupon_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    code = 'coupon_code'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/coupons/' + code, text=mock_response())
        response = client.coupons().destroy(code)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response.code == code


def test_invalid_destroy_coupon_request():
    client = Client(api_key='invalid')
    code = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/coupons/' + code, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.coupons().destroy(code)


def test_valid_find_all_coupon_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/coupons', text=mock_collection_response())
        response = client.coupons().find_all()

    assert response['coupons'][0].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1111'
    assert response['meta']['current_page'] == 1


def test_valid_find_all_coupon_request_with_options():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/coupons?per_page=2&page=1', text=mock_collection_response())
        response = client.coupons().find_all({'per_page': 2, 'page': 1})

    assert response['coupons'][1].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1222'
    assert response['meta']['current_page'] == 1


def test_invalid_find_all_coupon_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/coupons', status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.coupons().find_all()



import sys
from typing import Any, TypeVar
try:
    from functools import cached_property
except ImportError:
    cached_property = property  # type: ignore
try:
    from typing import ParamSpec
except ImportError:  # Python 3.7, Python 3.8, Python 3.9
    from typing_extensions import ParamSpec
if sys.version_info >= (3, 9):
    from collections.abc import Callable

T = TypeVar("T")
P = ParamSpec("P")


class Proxy(object):
    def __init__(self, _obj: object) -> None:
        object.__setattr__(self, "_obj", _obj)

    def __getattr__(self, name) -> Any:
        if name == "_obj":
            return object.__getattribute__(self, name)
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value) -> None:
        setattr(self._obj, name, value)

    def __call__(self) -> Any:
        return self._obj

    def __repr__(self) -> str:
        return repr(self._obj)

    def __str__(self) -> str:
        return str(self._obj)

    def __format__(self, format_spec: str) -> str:
        return format(self._obj, format_spec)

    def __hash__(self) -> int:
        return hash(self._obj)


if sys.version_info >= (3, 9):
    def callable_cached_property(func: Callable[P, T]) -> cached_property[T]:
        return cached_property(lambda s: Proxy(func(s)))  # type: ignore
else:
    def callable_cached_property(func):
        return cached_property(lambda s: Proxy(func(s)))



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import Organization, OrganizationBillingConfiguration


def organization_object():
    return Organization(
        webhook_url="https://test-example.example",
        billing_configuration=OrganizationBillingConfiguration(
            invoice_footer='footer',
            invoice_grace_period=3,
            vat_rate=20,
            document_locale="fr"
        )
    )


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/organization.json')

    with open(data_path, 'r') as organization_response:
        return organization_response.read()


def test_valid_update_organization_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/organizations',
                       text=mock_response())
        response = client.organizations().update(organization_object())

    assert response.name == 'Hooli'


def test_invalid_update_organization_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/organizations',
                       status_code=401,
                       text='')

        with pytest.raises(LagoApiError):
            client.organizations().update(organization_object())



from .applied_add_on import AppliedAddOn
from .applied_coupon import AppliedCoupon
from .billable_metric import BillableMetric, BillableMetricGroup
from .charge import Charge, Charges, ChargesResponse
from .coupon import Coupon, LimitationConfiguration
from .credit import CreditResponse, CreditsResponse
from .credit_note import Item, Items, CreditNote, CreditNoteUpdate
from .plan import Plan
from .add_on import AddOn
from .organization import Organization, OrganizationBillingConfiguration
from .event import Event, BatchEvent
from .fee import Fee
from .customer import Customer, CustomerBillingConfiguration, Metadata, MetadataList
from .invoice import InvoicePaymentStatusChange, Invoice, InvoiceMetadata, InvoiceMetadataList
from .invoice_item import InvoiceItemResponse
from .subscription import Subscription
from .customer_usage import Metric, ChargeObject, ChargeUsage, CustomerUsageResponse
from .wallet import Wallet
from .wallet_transaction import WalletTransaction



import sys
from typing import Union
try:
    from typing import Final
except ImportError:  # Python 3.7
    from typing_extensions import Final  # type: ignore
from urllib.parse import urljoin, urlencode

import requests

from ..version import LAGO_VERSION

if sys.version_info >= (3, 9):
    from collections.abc import Mapping, Sequence
else:
    from typing import Mapping, Sequence

URI_TEMPLATE: Final[str] = '{uri_path}{uri_query}'
QUERY_TEMPLATE: Final[str] = '?{query}'


def make_url(*, origin: str, path_parts: Sequence[str], query_pairs: Mapping[str, Union[int, str]] = {}) -> str:
    """Return url."""
    return urljoin(
        origin,
        URI_TEMPLATE.format(
            uri_path='/'.join(path_parts),
            uri_query=QUERY_TEMPLATE.format(
                query=urlencode(query_pairs),
            ) if query_pairs else '',
        ),
    )


def make_headers(*, api_key: str) -> Mapping[str, str]:
    """Return headers."""
    return {
        'Content-type': 'application/json',
        'Authorization': "Bearer {api_key}".format(api_key=api_key),
        'User-agent': 'Lago Python v{version}'.format(version=LAGO_VERSION),
    }


send_get_request = requests.get
send_post_request = requests.post
send_put_request = requests.put
send_delete_request = requests.delete



import sys

from lago_python_client.functools_ext import callable_cached_property


class InternalCollectionClient:
    def method(self) -> int:
        return 1


class Client:
    @callable_cached_property
    def collection(self) -> InternalCollectionClient:
        return InternalCollectionClient()


def test_callable_cached_property():
    # Given `InternalCollectionClient` class
    # And `Client` class with `callable_cached_property` decorator on `collection` method
    # And `client` instance
    client = Client()

    # When we call collection as a method
    method_result = client.collection().method()
    # Or as a property
    property_result = client.collection.method()

    # Then both results are equal
    assert property_result == method_result
    if sys.version_info >= (3, 8):
        # And collection objects are stored in cache during first request
        assert client.collection() == client.collection()
        assert client.collection == client.collection
        # ... but only for same `Client` instance
        assert Client().collection() != Client().collection()
        assert Client().collection != Client().collection



from typing import Optional

from pydantic import BaseModel

from .charge import Charges, ChargesResponse
from ..base_model import BaseResponseModel


class Plan(BaseModel):
    name: Optional[str]
    code: Optional[str]
    interval: Optional[str]
    description: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    trial_period: Optional[float]
    pay_in_advance: Optional[bool]
    bill_charges_monthly: Optional[bool]
    charges: Optional[Charges]


class PlanResponse(BaseResponseModel):
    lago_id: str
    name: str
    created_at: str
    code: str
    interval: Optional[str]
    description: Optional[str]
    amount_cents: Optional[int]
    amount_currency: Optional[str]
    trial_period: Optional[float]
    pay_in_advance: Optional[bool]
    bill_charges_monthly: Optional[bool]
    charges: Optional[ChargesResponse]
    active_subscriptions_count: int
    draft_invoices_count: int



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import Wallet


def wallet_object():
    return Wallet(
        name='name',
        external_customer_id='12345',
        rate_amount='1',
        paid_credits='10',
        granted_credits='10'
    )


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/wallet.json')

    with open(data_path, 'r') as wallet_response:
        return wallet_response.read()


def mock_collection_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/wallet_index.json')

    with open(data_path, 'r') as wallet_response:
        return wallet_response.read()


def test_valid_create_wallet_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/wallets', text=mock_response())
        response = client.wallets().create(wallet_object())

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'


def test_invalid_create_wallet_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/wallets', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.wallets().create(wallet_object())


def test_valid_update_wallet_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    arg = 'b7ab2926-1de8-4428-9bcd-779314ac129b'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/wallets/' + arg,
                       text=mock_response())
        response = client.wallets().update(wallet_object(), arg)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'


def test_invalid_update_wallet_request():
    client = Client(api_key='invalid')
    arg = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('PUT',
                       'https://api.getlago.com/api/v1/wallets/' + arg,
                       status_code=401,
                       text='')

        with pytest.raises(LagoApiError):
            client.wallets().update(wallet_object(), arg)


def test_valid_find_wallet_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    arg = 'b7ab2926-1de8-4428-9bcd-779314ac129b'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/wallets/' + arg, text=mock_response())
        response = client.wallets().find(arg)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'


def test_invalid_find_wallet_request():
    client = Client(api_key='invalid')
    arg = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/wallets/' + arg, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.wallets().find(arg)


def test_valid_destroy_wallet_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    arg = 'b7ab2926-1de8-4428-9bcd-779314ac129b'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/wallets/' + arg, text=mock_response())
        response = client.wallets().destroy(arg)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'


def test_invalid_destroy_wallet_request():
    client = Client(api_key='invalid')
    arg = 'invalid'

    with requests_mock.Mocker() as m:
        m.register_uri('DELETE', 'https://api.getlago.com/api/v1/wallets/' + arg, status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.wallets().destroy(arg)


def test_valid_find_all_wallet_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/wallets', text=mock_collection_response())
        response = client.wallets().find_all()

    assert response['wallets'][0].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response['meta']['current_page'] == 1


def test_valid_find_all_wallet_request_with_options():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/wallets?external_customer_id=123&per_page=2&page=1', text=mock_collection_response())
        response = client.wallets().find_all({'external_customer_id': 123, 'per_page': 2, 'page': 1})

    assert response['wallets'][1].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac1111'
    assert response['meta']['current_page'] == 1


def test_invalid_find_all_wallet_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/wallets', status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.wallets().find_all()



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin, DestroyCommandMixin, FindAllCommandMixin, FindCommandMixin
from ..models.customer import CustomerResponse
from ..models.customer_usage import CustomerUsageResponse
from ..services.json import from_json
from ..services.request import make_headers, make_url, send_get_request
from ..services.response import get_response_data, prepare_object_response, verify_response, Response


class CustomerClient(
    CreateCommandMixin[CustomerResponse],
    DestroyCommandMixin[CustomerResponse],
    FindAllCommandMixin[CustomerResponse],
    FindCommandMixin[CustomerResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'customers'
    RESPONSE_MODEL: ClassVar[Type[CustomerResponse]] = CustomerResponse
    ROOT_NAME: ClassVar[str] = 'customer'

    def current_usage(self, resource_id: str, external_subscription_id: str) -> CustomerUsageResponse:
        api_response: Response = send_get_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, resource_id, 'current_usage'),
                query_pairs={
                    'external_subscription_id': external_subscription_id,
                },
            ),
            headers=make_headers(api_key=self.api_key),
        )

        return prepare_object_response(
            response_model=CustomerUsageResponse,
            data=get_response_data(response=api_response, key='customer_usage'),
        )

    def portal_url(self, resource_id: str) -> str:
        api_response: Response = send_get_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, resource_id, 'portal_url'),
            ),
            headers=make_headers(api_key=self.api_key),
        )

        return get_response_data(response=api_response, key=self.ROOT_NAME).get('portal_url', '')



import os

import pytest
import requests_mock

from lago_python_client.client import Client
from lago_python_client.exceptions import LagoApiError
from lago_python_client.models import AppliedCoupon


def create_applied_coupon():
    return AppliedCoupon(
        external_customer_id='5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba',
        coupon_code='Free-Lemon-Juice'
    )


def mock_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/applied_coupon.json')

    with open(data_path, 'r') as applied_coupon_response:
        return applied_coupon_response.read()


def mock_collection_response():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_dir, 'fixtures/applied_coupon_index.json')

    with open(data_path, 'r') as applied_coupon_response:
        return applied_coupon_response.read()


def test_valid_create_applied_coupon_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/applied_coupons', text=mock_response())
        response = client.applied_coupons().create(create_applied_coupon())

    assert response.external_customer_id == '5eb02857-a71e-4ea2-bcf9-57d3a41bc6ba'


def test_invalid_create_applied_coupon_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('POST', 'https://api.getlago.com/api/v1/applied_coupons', status_code=401, text='')

        with pytest.raises(LagoApiError):
            client.applied_coupons().create(create_applied_coupon())


def test_valid_find_all_applied_coupon_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/applied_coupons', text=mock_collection_response())
        response = client.applied_coupons().find_all()

    assert response['applied_coupons'][0].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'
    assert response['meta']['current_page'] == 1


def test_valid_find_all_applied_coupon_request_with_options():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/applied_coupons?per_page=2&page=1', text=mock_collection_response())
        response = client.applied_coupons().find_all({'per_page': 2, 'page': 1})

    assert response['applied_coupons'][1].lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac2222'
    assert response['meta']['current_page'] == 1


def test_invalid_find_all_applied_coupon_request():
    client = Client(api_key='invalid')

    with requests_mock.Mocker() as m:
        m.register_uri('GET', 'https://api.getlago.com/api/v1/applied_coupons', status_code=404, text='')

        with pytest.raises(LagoApiError):
            client.applied_coupons().find_all()


def test_valid_destroy_applied_coupon_request():
    client = Client(api_key='886fe239-927d-4072-ab72-6dd345e8dd0d')
    external_customer_id = 'external_customer_id'
    applied_coupon_id = '_ID_'

    with requests_mock.Mocker() as m:
        m.register_uri(
            'DELETE',
            'https://api.getlago.com/api/v1/customers/' + external_customer_id + '/applied_coupons/' + applied_coupon_id,
            text=mock_response()
        )
        response = client.applied_coupons().destroy(external_customer_id, applied_coupon_id)

    assert response.lago_id == 'b7ab2926-1de8-4428-9bcd-779314ac129b'


def test_invalid_destroy_applied_coupon_request():
    client = Client(api_key='invalid')
    external_customer_id = 'external_customer_id'
    applied_coupon_id = '_ID_'

    with requests_mock.Mocker() as m:
        m.register_uri(
            'DELETE',
            'https://api.getlago.com/api/v1/customers/' + external_customer_id + '/applied_coupons/' + applied_coupon_id,
            status_code=404,
            text=''
        )
        with pytest.raises(LagoApiError):
            client.applied_coupons().destroy(external_customer_id, applied_coupon_id)



"""Test JSON services."""
import pytest
from requests import Response

from lago_python_client.exceptions import LagoApiError
from lago_python_client.services.json import from_json, to_json


def test_from_json():
    """Deserialize data from json string."""
    # Given some data
    expected_data = {'a': {'b': 'c'}}
    # And same data serialized in json format stored as string,
    json_string = '{"a":{"b":"c"}}'
    # ... or bytes,
    json_bytes = b'{"a":{"b":"c"}}'
    # ... or even available as ``content`` property of instanse of ``requests.Response`` class
    json_requests_response = Response()
    json_requests_response._content_consumed = True
    json_requests_response._content = json_bytes
    json_requests_response.status_code = 200
    # ... or None
    json_none = None
    # ... or something incorrect
    json_string_shit_happens = b'{"abc'

    # When service is applied
    # Then service result is equal to given data.
    assert from_json(json_string) == expected_data
    assert from_json(json_bytes) == expected_data
    assert from_json(json_requests_response) == expected_data
    # ... or raise exception
    with pytest.raises(LagoApiError) as cm:
        from_json(json_none)
    assert cm.value.detail == 'Input must be bytes, bytearray, memoryview, or str'
    with pytest.raises(LagoApiError):
        from_json(json_string_shit_happens)


def test_to_json():
    """Serialize data to json string."""
    # Given some data
    initial_data = {'a': {'b': 'c'}}
    # And same data serialized in json format
    expected_json_string = '{"a":{"b":"c"}}'

    # When service is applied
    # Then service result is equal to given serialized json string.
    assert to_json(initial_data) == expected_json_string



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin, DestroyCommandMixin, FindAllCommandMixin, FindCommandMixin, UpdateCommandMixin
from ..models.plan import PlanResponse


class PlanClient(
    CreateCommandMixin[PlanResponse],
    DestroyCommandMixin[PlanResponse],
    FindAllCommandMixin[PlanResponse],
    FindCommandMixin[PlanResponse],
    UpdateCommandMixin[PlanResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'plans'
    RESPONSE_MODEL: ClassVar[Type[PlanResponse]] = PlanResponse
    ROOT_NAME: ClassVar[str] = 'plan'



import sys
from typing import Any, ClassVar, Optional, Type

from pydantic import BaseModel

from .base_client import BaseClient
from .fee_client import FeeClient
from ..mixins import CreateCommandMixin, FindCommandMixin
from ..models.event import EventResponse
from ..models.fee import FeeResponse
from ..services.json import to_json
from ..services.request import make_headers, make_url, send_post_request
from ..services.response import get_response_data, prepare_object_list_response, verify_response, Response


if sys.version_info >= (3, 9):
    from collections.abc import Mapping
else:
    from typing import Mapping

class EventClient(
    CreateCommandMixin[EventResponse],
    FindCommandMixin[EventResponse],
    BaseClient
):
    API_RESOURCE: ClassVar[str] = 'events'
    RESPONSE_MODEL: ClassVar[Type[EventResponse]] = EventResponse
    ROOT_NAME: ClassVar[str] = 'event'

    def batch_create(self, input_object: BaseModel) -> Optional[bool]:
        api_response: Response = send_post_request(
            url=make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, 'batch'),
            ),
            data=to_json({
                self.ROOT_NAME: input_object.dict()
            }),
            headers=make_headers(api_key=self.api_key),
        )
        verify_response(api_response)

        return True # TODO: should return None

    def estimate_fees(self, input_object: BaseModel) -> Mapping[str, Any]:
        api_response: Response = send_post_request(
            url= make_url(
                origin=self.base_url,
                path_parts=(self.API_RESOURCE, 'estimate_fees'),
            ),
            data=to_json({
                self.ROOT_NAME: input_object.dict()
            }),
            headers=make_headers(api_key=self.api_key),
        )

        return prepare_object_list_response(
            api_resource=FeeClient.API_RESOURCE,
            response_model=FeeResponse,
            data=get_response_data(response=api_response, key='fees'),
        )



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin, DestroyCommandMixin, FindAllCommandMixin, FindCommandMixin, UpdateCommandMixin
from ..models.billable_metric import BillableMetricResponse


class BillableMetricClient(
    CreateCommandMixin[BillableMetricResponse],
    DestroyCommandMixin[BillableMetricResponse],
    FindAllCommandMixin[BillableMetricResponse],
    FindCommandMixin[BillableMetricResponse],
    UpdateCommandMixin[BillableMetricResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'billable_metrics'
    RESPONSE_MODEL: ClassVar[Type[BillableMetricResponse]] = BillableMetricResponse
    ROOT_NAME: ClassVar[str] = 'billable_metric'



import sys
from typing import Any, ClassVar, Type, Union

from .base_client import BaseClient
from ..models.group import GroupResponse
from ..services.request import make_headers, make_url, send_get_request
from ..services.response import get_response_data, prepare_index_response, Response

if sys.version_info >= (3, 9):
    from collections.abc import Mapping
else:
    from typing import Mapping


class GroupClient(
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'groups'
    RESPONSE_MODEL: ClassVar[Type[GroupResponse]] = GroupResponse
    ROOT_NAME: ClassVar[str] = 'group'

    def find_all(self, metric_code: str, options: Mapping[str, Union[int, str]] = {}) -> Mapping[str, Any]:
        api_response: Response = send_get_request(
            url=make_url(
                origin=self.base_url,
                path_parts=('billable_metrics', metric_code, self.API_RESOURCE),
                query_pairs=options,
            ),
            headers=make_headers(api_key=self.api_key),
        )

        return prepare_index_response(
            api_resource=self.API_RESOURCE,
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response),
        )



from http import HTTPStatus
import sys
from typing import Any, Optional

if sys.version_info < (3, 9):
    from typing import MutableMapping
else:
    from collections.abc import MutableMapping


class LagoApiError(Exception):
    def __init__(
        self,
        status_code: int,
        url: Optional[str],
        response: Any,
        detail: Optional[str] = None,
        headers: Optional[MutableMapping[str, str]] = None,
    ) -> None:
        if detail is None:
            detail = HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.url = url
        self.response = response
        self.detail = detail
        self.headers = headers

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin
from ..models.applied_add_on import AppliedAddOnResponse


class AppliedAddOnClient(
    CreateCommandMixin[AppliedAddOnResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'applied_add_ons'
    RESPONSE_MODEL: ClassVar[Type[AppliedAddOnResponse]] = AppliedAddOnResponse
    ROOT_NAME: ClassVar[str] = 'applied_add_on'



from typing import ClassVar, Type

from .base_client import BaseClient
from ..mixins import CreateCommandMixin, FindAllCommandMixin
from ..models.applied_coupon import AppliedCouponResponse
from ..services.request import make_headers, make_url, send_delete_request
from ..services.response import get_response_data, prepare_object_response, Response


class AppliedCouponClient(
    CreateCommandMixin[AppliedCouponResponse],
    FindAllCommandMixin[AppliedCouponResponse],
    BaseClient,
):
    API_RESOURCE: ClassVar[str] = 'applied_coupons'
    RESPONSE_MODEL: ClassVar[Type[AppliedCouponResponse]] = AppliedCouponResponse
    ROOT_NAME: ClassVar[str] = 'applied_coupon'

    def destroy(self, external_customer_id: str, applied_coupon_id: str) -> AppliedCouponResponse:
        api_response: Response = send_delete_request(
            url=make_url(
                origin=self.base_url,
                path_parts=('customers', external_customer_id, self.API_RESOURCE, applied_coupon_id),
            ),
            headers=make_headers(api_key=self.api_key),
        )

        return prepare_object_response(
            response_model=self.RESPONSE_MODEL,
            data=get_response_data(response=api_response, key=self.ROOT_NAME),
        )



from typing import Optional, List

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class OrganizationBillingConfiguration(BaseModel):
    invoice_footer: Optional[str]
    invoice_grace_period: Optional[int]
    vat_rate: Optional[float]
    document_locale: Optional[str]


class Organization(BaseModel):
    webhook_url: Optional[str]
    country: Optional[str]
    address_line1: Optional[str]
    address_line2: Optional[str]
    state: Optional[str]
    zipcode: Optional[str]
    email: Optional[str]
    city: Optional[str]
    legal_name: Optional[str]
    legal_number: Optional[str]
    timezone: Optional[str]
    email_settings: Optional[List[str]]
    billing_configuration: Optional[OrganizationBillingConfiguration]


class OrganizationResponse(BaseResponseModel):
    name: str
    created_at: str
    webhook_url: Optional[str]
    country: Optional[str]
    address_line1: Optional[str]
    address_line2: Optional[str]
    state: Optional[str]
    zipcode: Optional[str]
    email: Optional[str]
    city: Optional[str]
    legal_name: Optional[str]
    legal_number: Optional[str]
    timezone: Optional[str]
    email_settings: Optional[List[str]]
    billing_configuration: Optional[OrganizationBillingConfiguration]



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from ..base_model import BaseResponseModel


class BillableMetricGroup(BaseModel):
    key: Optional[str]
    values: Optional[List[Union[str, Dict[str, Any]]]]


class BillableMetric(BaseModel):
    name: Optional[str]
    code: Optional[str]
    description: Optional[str]
    aggregation_type: Optional[str]
    field_name: Optional[str]
    group: Optional[BillableMetricGroup]


class BillableMetricResponse(BaseResponseModel):
    lago_id: str
    name: str
    code: str
    description: Optional[str]
    aggregation_type: Optional[str]
    field_name: Optional[str]
    created_at: str
    group: BillableMetricGroup
    active_subscriptions_count: int
    draft_invoices_count: int
    plans_count: int
